import torch
import numpy as np
from PIL import Image
import os
import datetime

from model import*


epoch_num = 20
batch_size = 15
chanel_num = 3
image_file_num = 6
block_num = 10
block_set_num = image_file_num*block_num
block_size = 64
train_block_size = 60
light_kind_num = 6
motion_num = 25
train_data_num = chanel_num*block_set_num*light_kind_num*motion_num
check_num = train_data_num
check_num = check_num - check_num%15

shots_num = 3
view_point_row_num = 5
view_point_num = view_point_row_num*view_point_row_num
dev = 0.005

aug_data = np.zeros((batch_size, view_point_num, block_size, block_size), np.float32)
aug_data_frame1 = np.zeros((batch_size, view_point_num, train_block_size, train_block_size), np.float32)
aug_data_frame2 = np.zeros((batch_size, view_point_num, train_block_size, train_block_size), np.float32)
aug_data_frame3 = np.zeros((batch_size, view_point_num, train_block_size, train_block_size), np.float32)

cnt = 0
loss_sum = 0.0

device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")


if os.path.isdir("./Dst") == False:
	os.mkdir("./Dst")

if os.path.isdir("./Dst/Mask") == False:
	os.mkdir("./Dst/Mask")

if os.path.isdir("./Dst/Model") == False:
	os.mkdir("./Dst/Model")




#Load train datas
load_light_field = np.zeros((block_set_num*chanel_num, view_point_num, block_size, block_size), np.uint8)
start_index_R = 0;
start_index_G = image_file_num*block_num
start_index_B = image_file_num*block_num*2

print("train with %f devietion noise" %dev)
for n in range(image_file_num):
	if(n % 10 == 0):
		print("load %04d" % (n))

	fname = "./TrainingData/%03d.png" % (n)
	img_load = Image.open(fname)
	img_load = np.asarray(img_load)

	for b in range(block_num):
		for t in range(view_point_num):
			load_light_field[n*block_num+b +start_index_R,t,:,:] = img_load[block_size*b:block_size*b+block_size,block_size*t:block_size*t+block_size,0]
			load_light_field[n*block_num+b +start_index_G,t,:,:] = img_load[block_size*b:block_size*b+block_size,block_size*t:block_size*t+block_size,1]
			load_light_field[n*block_num+b +start_index_B,t,:,:] = img_load[block_size*b:block_size*b+block_size,block_size*t:block_size*t+block_size,2]





#Model initaition
shot_A = Shot()
shot_A.to(device)
shot_A_optimizer = torch.optim.Adam(shot_A.parameters(), lr=0.0001, betas=(0.9, 0.999), eps=1e-08)

shot_B = Shot()
shot_B.to(device)
shot_B_optimizer = torch.optim.Adam(shot_B.parameters(), lr=0.0001, betas=(0.9, 0.999), eps=1e-08)

reconstruct = Reconstruct()
reconstruct.to(device)
reconstruct_optimizer = torch.optim.Adam(reconstruct.parameters(), lr=0.0001, betas=(0.9, 0.999), eps=1e-08)

vdsr = VeryDeepSuperResolution()
vdsr.to(device)
vdsr_optimizer = torch.optim.Adam(vdsr.parameters(), lr=0.0001, betas=(0.9, 0.999), eps=1e-08)


#Physical limitations of masks
shot_A.cn01.weight.data[shot_A.cn01.weight.data < 0.0] = 0.0
shot_A.cn01.weight.data[shot_A.cn01.weight.data > 1.0] = 1.0
shot_B.cn01.weight.data[shot_B.cn01.weight.data < 0.0] = 0.0
shot_B.cn01.weight.data[shot_B.cn01.weight.data > 1.0] = 1.0



#Train loop
for i in range(1, epoch_num+1):

	sff = np.random.permutation(train_data_num)

	for n in range(0, train_data_num, batch_size):
		#Data augmentation by intensity levels
		for j in range(batch_size):
			aug_data[j,:,:,:] = ((1-(int(sff[n+j]/motion_num)%light_kind_num)/10)*load_light_field[int(sff[n+j]/light_kind_num/motion_num)]/255).astype(np.float32)

            #Creating pseudo motion
			shift_x = int((sff[n+j]%motion_num)%5) -2
			shift_y = int((sff[n+j]%motion_num)/5) -2
			aug_data_frame1[j,:,:,:]= aug_data[j, : , (2-shift_x):(2+train_block_size-shift_x), (2-shift_y):(2+train_block_size-shift_y)]
			aug_data_frame2[j,:,:,:]= aug_data[j, : , 2:2+train_block_size, 2:2+train_block_size]
			aug_data_frame3[j,:,:,:]= aug_data[j, : , (2+shift_x):(2+train_block_size+shift_x), (2+shift_y):(2+train_block_size+shift_y)]


		gray_light_field_frame1 = torch.from_numpy(aug_data_frame1).clone().to(device)
		gray_light_field_frame2 = torch.from_numpy(aug_data_frame2).clone().to(device)
		gray_light_field_frame3 = torch.from_numpy(aug_data_frame3).clone().to(device)


		#Train models
		shot_A.zero_grad()
		shot_B.zero_grad()
		reconstruct.zero_grad()
		vdsr.zero_grad()


		gray_acquired_image1A = shot_A(gray_light_field_frame1)
		gray_acquired_image2B = shot_B(gray_light_field_frame2)
		gray_acquired_image3A = shot_A(gray_light_field_frame3)

		gray_acquired_images_ABA = torch.cat((gray_acquired_image1A, gray_acquired_image2B, gray_acquired_image3A), dim=1)
		gray_noise_acquired_images_ABA = gray_acquired_images_ABA + view_point_num*dev*torch.randn(batch_size,shots_num,train_block_size,train_block_size).to(device)

		gray_makeshift_light_field = reconstruct(gray_noise_acquired_images_ABA)
		gray_reconstruct_light_field = vdsr(gray_makeshift_light_field)

		loss = F.mse_loss(gray_reconstruct_light_field, gray_light_field_frame2)

		loss.backward()
		shot_A_optimizer.step()
		shot_B_optimizer.step()
		reconstruct_optimizer.step()
		vdsr_optimizer.step()

		shot_A.cn01.weight.data[shot_A.cn01.weight.data < 0.0] = 0.0
		shot_A.cn01.weight.data[shot_A.cn01.weight.data > 1.0] = 1.0
		shot_B.cn01.weight.data[shot_B.cn01.weight.data < 0.0] = 0.0
		shot_B.cn01.weight.data[shot_B.cn01.weight.data > 1.0] = 1.0

		# print("epoch:%04d/%04d, data:%06d/%06d, psnr:%.6f" % (i, epoch_num, n, train_data_num, 10.0*torch.log10(1.0/loss.data)))

		loss_data = loss.data.to("cpu")
		set_num = int((n / train_data_num * 100.) // 10)
		pro_bar = ('=' * set_num) + (' ' * (10 - set_num))
		# print("epoch:%04d/%04d, data:%06d/%06d, psnr:%.6f" % (i, epoch_num, n, train_data_num, 10*np.log10(1.0/loss_data)))
		print("\rEpoch:{:04d}/{:04d}, Data:{:06d}/{:06d} [{}], PSNR:{:08.6f}".format(i, epoch_num, n, train_data_num, pro_bar, 10*np.log10(1.0/loss_data)), end='')

		
		loss_sum += loss.data
		cnt += 1

		#Save masks and calculate average psnr
		if(n%check_num == 0):

			mask_txt_name = "./Dst/Mask/mask_txt_%03depoch-%06ddata.txt" % (i, n)
			f_mask = open(mask_txt_name, "w")

			img_save = shot_A.cn01.weight.data[0,:,0,0].to("cpu")
			img_save = img_save.reshape(view_point_row_num,view_point_row_num)
			f_mask.write("%s\n" % img_save)
			img_save = Image.fromarray(np.uint8(img_save * 255.0))
			img_save = img_save.resize((256,256), Image.NEAREST)
			img_save.save("./Dst/Mask/Mask_A_epoch{:04d}-batch{:06d}done.png".format(i, n))

			img_save = shot_B.cn01.weight.data[0,:,0,0].to("cpu")
			img_save = img_save.reshape(view_point_row_num,view_point_row_num)
			f_mask.write("%s\n" % img_save)
			img_save = Image.fromarray(np.uint8(img_save * 255.0))
			img_save = img_save.resize((256,256), Image.NEAREST)
			img_save.save("./Dst/Mask/Mask_B_epoch{:04d}-batch{:06d}done.png".format(i, n))

			f_mask.close()


			print("\n@@@    average psnr:%.6f    @@@" % (10.0*torch.log10(1.0 / (loss_sum / cnt))))
			f_psnr = open("./Dst/average_psnr.txt", "a")
			f_psnr.write("epoch:%04d/%04d, data:%06d/%06d, psnr:%.6f\n" % (i, epoch_num, n, train_data_num, 10.0*torch.log10(1.0 / (loss_sum / cnt))))
			f_psnr.close()

			cnt = 0
			loss_sum = 0.0



	#Save models every epoch
	shot_A.to("cpu")
	torch.save(shot_A.state_dict(), "./Dst/Model/shot_A%04d.pth" %i)
	shot_A.to(device)

	shot_B.to("cpu")
	torch.save(shot_B.state_dict(), "./Dst/Model/shot_B%04d.pth" %i)
	shot_B.to(device)

	reconstruct.to("cpu")
	torch.save(reconstruct.state_dict(), "./Dst/Model/reconstruct%04d.pth" %i)
	reconstruct.to(device)

	vdsr.to("cpu")
	torch.save(vdsr.state_dict(), "./Dst/Model/vdsr%04d.pth" %i)
	vdsr.to(device)
